/*
 * himwidget.cc
 * 
 * History Image Viewer Widget implementation. 
 * 
 * Copyright (c) 2004 by Wolfgang Wieser (wwieser@gmx.de)
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. 
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#include "himview.h"

#include <qfiledialog.h>
#include <qmessagebox.h>
#include <qcursor.h>

#include <string.h>


int HistoryViewWidget::n_open_windows=0;


void HistoryViewWidget::_DisplayCaption()
{
	int idx=-1;
	for(HEntry *i=active; i; i=i->next)  ++idx;
	
	QString tmp;
	const char *f=strrchr(active->file,'/');
	if(f)  ++f;  else  f=active->file;
	tmp.sprintf("%s [%d/%d/%d]",f,idx,hl.count(),hist_size);
	
	setCaption(tmp);
}


void HistoryViewWidget::_DisplayActiveImage()
{
	if(!active)
	{
		erase();
		setCaption("[no active image]");
		// This is a hack; I want to get rid of fixed size somehow: 
		setMinimumSize(QSize(1,1));
		setMaximumSize(QSize(32767,32767));
	}
	else
	{
		_DisplayCaption();
		
		QSize size(active->img->width(),active->img->height());
		resize(size);
		setFixedSize(size);
		//raise();
		show();
		
		pnt.begin(this);
		pnt.drawImage(QPoint(0,0),*active->img,active->img->rect());
		pnt.end();
	}
}


void HistoryViewWidget::_CheckExpire()
{
	// Check if old images expire. 
	bool active_deleted=0;
	for(int cnt=hl.count(); cnt>hist_size; cnt--)
	{
		HEntry *ent=hl.first();
		if(active==ent)
		{  active=active->next;  active_deleted=1;  }
		delete hl.popfirst();
	}
	
	if(active_deleted)
	{  _DisplayActiveImage();  }
}


int HistoryViewWidget::LoadFile(const QString &file)
{
	if(file.isEmpty())  return(1);
	
	// Load new entry and put it at the end: 
	HEntry *ent=new HEntry(file);
	
	if(!ent->img)
	{
		// Loading failed. 
		QMessageBox::warning(this,"Loading failed",
			"Failed to load image file\n" + file + ".",
			QMessageBox::Ok,QMessageBox::NoButton);
		delete ent;
		return(1);
	}
	
	hl.append(ent);
	active=ent;
	_CheckExpire();
	
	_DisplayActiveImage();
	
	return(0);
}


void HistoryViewWidget::Reload(bool move_hist,bool atend)
{
	if(!active)  return;
	
	HEntry *ent=new HEntry(active->file);
	if(!ent->img)
	{
		// Loading failed. 
		QMessageBox::warning(this,"Loading failed",
			"Failed to reload image file\n" + active->file + ".",
			QMessageBox::Ok,QMessageBox::NoButton);
		delete ent;
		return;
	}
	
	if(atend)
	{
		// Queue at the end: 
		hl.append(ent);
		
		// If move_hist is set, jump to the end: 
		if(move_hist)
		{  active=ent;  }
		
		_CheckExpire();
	}
	else
	{
		// Queue after the current image and make it active: 
		hl.queueafter(ent,active);
		HEntry *prev=active;
		active=ent;
		
		// If move_hist is not set, delete the old image so that the 
		// new one replaces the old one. 
		if(!move_hist)
		{
			hl.dequeue(prev);
			delete prev;
		}
		else
		{  _CheckExpire();  }
	}
	
	_DisplayActiveImage();
}


void HistoryViewWidget::DeleteImage(bool all)
{
	if(all)
	{
		while(!hl.is_empty())
		{  delete hl.popfirst();  }
		active=NULL;
	}
	else if(active)
	{
		HEntry *ent=active;
		if(active->next)  active=active->next;
		else if(active->prev)  active=active->prev;
		else active=NULL;
		
		hl.dequeue(ent);
		delete ent;
	}
	
	_DisplayActiveImage();
}


void HistoryViewWidget::_SwitchHist(int dir)
{
	if(dir<0)
	{
		// Go back in history: 
		if(active && active->prev)
		{  active=active->prev;  _DisplayActiveImage();  }
	}
	else
	{
		// Go forward in history: 
		if(active && active->next)
		{  active=active->next;  _DisplayActiveImage();  }
	}
}


void HistoryViewWidget::_NewCopyImage()
{
	HistoryViewWidget *hv=new HistoryViewWidget(
		active ? active->file : QString::null);
	hv->show();
}


void HistoryViewWidget::_ImageOpen()
{
#if 0
	QString file=QFileDialog::getOpenFileName(
		/*startWith=*/active ? active->file : QString::null,
		/*filter=*/"Images (*.png *.jpg)",
		/*parent=*/this,
		/*name=*/"Open file dialog",
		/*caption=*/"Choose a file to open");
	if(!file.isEmpty())
	{  LoadFile(file);  }
#else
	QStringList files=QFileDialog::getOpenFileNames(
		/*filter=*/"Images (*.png *.jpg)",
		/*dir=*/active ? active->file : QString::null,
		/*parent=*/this,
		/*name=*/"Open file dialog",
		/*caption=*/"Choose a file to open");
	for(QStringList::Iterator i=files.begin(); i!=files.end(); ++i)
	{  LoadFile(*i);  }
#endif	
}


void HistoryViewWidget::_ImageSaveAs()
{
	if(!active)  return;
	
	QString file=QFileDialog::getSaveFileName(
		/*startWith=*/active ? active->file : QString::null,
		/*filter=*/"Images (*.png *.jpg)",
		/*parent=*/this,
		/*name=*/"Save file dialog",
		/*caption=*/"Enter a file name");
	if(file.isEmpty())  return;
	
	int idx=-1;
	for(HEntry *i=active; i; i=i->next)  ++idx;
	
	const char *format="PNG";
	int qual=0;  // compress best
	if(file.right(4)==".jpg" || file.right(4)==".JPG")
	{  format="JPEG";  qual=75;  }
	
	if(!active->img->save(file,format,qual))
	{
		QString tmp;
		tmp.sprintf("Failed to save %s [%d]\nas %s (%s %d)",
			(const char*)active->file,idx,
			(const char*)file,
			format,qual);
		QMessageBox::warning(this,"Loading failed",tmp,
			QMessageBox::Ok,QMessageBox::NoButton);
	}
}


void HistoryViewWidget::Close()
{
	close();   // will delete us via WDestructiveClose. 
}

void HistoryViewWidget::closeEvent(QCloseEvent *ev)
{
	ev->accept();
}


void HistoryViewWidget::paintEvent(QPaintEvent *ev)
{
	if(active)
	{
		pnt.begin(this);
		
		//pnt.drawImage(QPoint(imgx,imgy),*active->img,active->img->rect());
		
		QRect rect=ev->rect();
		pnt.drawImage(QPoint(rect.x(),rect.y()),*active->img,rect);
		
		pnt.end();
	}
}


void HistoryViewWidget::keyPressEvent(QKeyEvent *ev)
{
	ev->accept();
	
	bool shift=ev->state() & ShiftButton;
	
	switch(ev->key())
	{
		case Key_Q:
			if(shift)  qapp->quit();
			else  Close();
			break;
		case Key_R:  Reload(shift ? 0 : 1,0);  break;
		case Key_A:  Reload(shift ? 0 : 1,1);  break;
		case Key_P:
		case Key_Prior:
			_SwitchHist(-1);  break;
		case Key_N:
		case Key_Next:
			_SwitchHist(+1);  break;
		case Key_D:  DeleteImage(shift);  break;
		case Key_C:  _NewCopyImage();  break;
		case Key_O:  _ImageOpen();  break;
		case Key_S:  _ImageSaveAs();  break;
		case Key_H:
		{
			QString text=
				" h\t(help) \tprint this help\n"
				" q\t(quit) \tclose window, quit if all windows are closed\n"
				" Q\t(quit) \tclose all windows and quit\n"
				" a\t(append) \tlike reload but at put image at the end and jump there\n"
				" A\t(append) \tlike 'a' but do not jump to end\n"
				" r\t(reload) \treload image, new image after current one\n"
				" R\t(replace) \treload image and replace current image\n"
				" p,n\t(prev,next) \tprev, next image in history\n"
				" d\t(delete) \tdelete current image\n"
				" D\t(delete) \tdelete all images\n"
				" o\t(open) \topen a file (opens select dialog)\n"
				" s\t(save) \tsave currently displayed image to file\n"
				" c\t(copy) \topen a new window with active image\n"
				" +/-\t\tincrease/decrease history size (see window caption)\n";
			QMessageBox::information(this,"Command buttons",
				text,"Hide");
		}	break;
		case Key_Plus:
			++hist_size;
			_DisplayCaption();
			break;
		case Key_Minus:
			if(hist_size>1)
			{
				--hist_size;
				_CheckExpire();
				_DisplayCaption();
			}
			break;
	}
}


void HistoryViewWidget::mousePressEvent(QMouseEvent *ev)
{
	if(ev->button()==RightButton)
	{
		_CreateMenu();
		
		int rv=menu->exec(QCursor::pos());
		switch(rv)
		{
			case -1:  break;  // cancelled
			case 'q':  Close();  break;
			case 'Q':  qapp->quit();  break;
			case 'a':  Reload(1,1);  break;
			case 'r':  Reload(1,0);  break;
			case 'R':  Reload(0,0);  break;
			case 'p':  _SwitchHist(-1);  break;
			case 'n':  _SwitchHist(+1);  break;
			case 'o':  _ImageOpen();  break;
			case 's':  _ImageSaveAs();  break;
			case 'c':  _NewCopyImage();  break;
			case 'd':  DeleteImage(0);  break;
		}
		if(rv>=256)
		{
			int idx=rv-256;
			HEntry *i=hl.last();
			for(; i && idx; i=i->prev)  --idx;
			if(i)
			{
				active=i;
				_DisplayActiveImage();
			}
		}
	}
}


void HistoryViewWidget::_CreateMenu()
{
	if(!menu)
	{
		menu=new QPopupMenu(this);
		
		switch_to_menu=new QPopupMenu(menu);
		switch_to_menu->setCheckable(1);
		
		menu->setCheckable(0);
		menu->insertItem("Append",'a');
		menu->insertItem("Reload",'r');
		menu->insertItem("Replace",'R');
		menu->insertItem("Delete",'d');
		menu->insertSeparator();
		menu->insertItem("Open...",'o');
		menu->insertItem("Save as...",'s');
		menu->insertItem("Copy into new",'c');
		menu->insertItem("Close",'q');
		menu->insertItem("Quit",'Q');
		menu->insertSeparator();
		menu->insertItem("Previous",'p');
		menu->insertItem("Next",'n');
		menu->insertItem("Switch to",switch_to_menu);
		//menu->insertSeparator();
	}
	
	switch_to_menu->clear();
	int idx=0;
	for(HEntry *i=hl.last(); i; i=i->prev,idx++)
	{
		const char *name=strrchr(i->file,'/');
		if(name)  ++name;  else  name=i->file;
		QString tmp;
		tmp.sprintf("%s [%d]",name,idx);
		switch_to_menu->insertItem(tmp,256+idx);
		if(i==active)
		{  switch_to_menu->setItemChecked(256+idx,1);  }
	}
	
	menu->setItemEnabled('a',active);
	menu->setItemEnabled('r',active);
	menu->setItemEnabled('R',active);
	menu->setItemEnabled('d',active);
	menu->setItemEnabled('s',active);
	menu->setItemEnabled('p',active && active->prev);
	menu->setItemEnabled('n',active && active->next);
}


HistoryViewWidget::HistoryViewWidget(const QString &file,
	QWidget *parent, const char *name,WFlags f) : 
	QWidget(parent,name,f|WDestructiveClose),
	pnt(),
	hl(),
	active(NULL),
	hist_size(def_hist_size)
{
	menu=NULL;
	switch_to_menu=NULL;
	
	//setBackgroundMode(FixedColor);
	setBackgroundColor(Qt::black);
	
	//setCursor(BlankCursor);
	
	if(file.isEmpty())
	{
		// No file specified. 
		// Set initial size to something reasonable: 
		resize(320,240);
	}
	else
	{
		LoadFile(file);
	}
	
	++n_open_windows;
}

HistoryViewWidget::~HistoryViewWidget()
{
	// Clean up history entries: 
	while(!hl.is_empty())
	{  delete hl.popfirst();  }
	
	if(!--n_open_windows)
	{
		qapp->quit();
	}
}

//------------------------------------------------------------------------------

HistoryViewWidget::HEntry::HEntry(const QString &_file) : 
	LinkedListBase<HEntry>(),
	file(_file)
{
	//fprintf(stderr,"Loading %s... ",(const char*)_file);
	
	img=new QImage(file);
	if(img->isNull())
	{  delete img;  img=NULL;  }
	
	//if(img)  fprintf(stderr,"OK\n");
	//else  fprintf(stderr,"failed\n");
}

HistoryViewWidget::HEntry::~HEntry()
{
	if(img)
	{  delete img;  img=NULL;  }
}
